// Auto-generated by cmake

#ifndef HIP_VERSION_H
#define HIP_VERSION_H

#define HIP_VERSION_MAJOR 6
#define HIP_VERSION_MINOR 2
#define HIP_VERSION_PATCH 41134
#define HIP_VERSION_GITHASH "65d174c3e"
#define HIP_VERSION_BUILD_ID 0
#define HIP_VERSION_BUILD_NAME ""
#define HIP_VERSION    (HIP_VERSION_MAJOR * 10000000 + HIP_VERSION_MINOR * 100000 + HIP_VERSION_PATCH)

#define __HIP_HAS_GET_PCH 1

#endif

